from __future__ import print_function
from future.utils import raise_


import os
import tempfile
import shutil
import shlex
import subprocess
import sys

from . import pipeline
from . import analysis
from . import util
from . import config

def print_lines(lines):
    print('\n'.join(lines))

def process_file(fname):
    basedir, name = os.path.split(fname)
    with open(fname) as f:
        lines = f.readlines()

    lines = pipeline.pipeline(lines, basedir)

    # TODO:
    # extract exports to fname.h
    # print("\n".join(exports))
    # print("/* END HEADER */")

    return lines

def run_cmd(cmd, more_args=[], stdin=None):
    cmd_args = shlex.split(cmd)
    cmd_args.extend(more_args)
    util.debug(" ".join(cmd_args))

    if stdin:
        pipe = subprocess.Popen(cmd_args, stdin=subprocess.PIPE,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = pipe.communicate(bytes(stdin))

        if stderr:
            print(stderr, file=sys.stderr)
        return stdout or ""

    return subprocess.check_output(cmd_args)

def process_h_file(tmp_dir, arg):
    name = arg
    fname = os.path.join(tmp_dir, arg)
    basedir = os.path.dirname(fname)

    try:
        os.makedirs(basedir)
    except:
        pass

    with open(arg) as f:
        lines = f.readlines()

    with open(fname, "w") as f:
        f.write("\n".join(lines))

    return

def print_file_with_line_nums(fname):
    util.debug("")
    util.debug('// ' + fname)
    with open(fname) as f:
        lines = f.readlines()

    for i, line in enumerate(lines):
        util.debug(i, line.rstrip())
    util.debug("")

def compile_cpy_file(tmp_dir, arg):
    name = arg.rstrip(".cpy")
    fname = os.path.join(tmp_dir, "%s.cpp" % name)
    return compile_cpp_file(tmp_dir, fname)

def compile_cpp_file(tmp_dir, arg):
    name = arg.rstrip(".cpp")
    fname = os.path.join(tmp_dir, "%s.cpp" % name)
    ofname = os.path.join(tmp_dir, "%s.o" % name)

    try:
        run_cmd("g++ -c '%s' -o '%s' " % (fname, ofname), COMPILE_FLAGS)
    except:
        if config.PRINT_ON_ERROR:
            print_file_with_line_nums(fname)

        e = sys.exc_info()
        print("Couldn't compile", fname, "aborting")
        sys.exit(1)
    return ofname


def add_guards(arg, lines):
    arg = arg.replace('/', '__')
    arg = arg.replace('.', '_')
    arg = arg.replace(' ', '_')
    arg = arg.upper()
    arg = "%s_H" % (arg)

    guard_line = "#ifndef %s\n#define %s" % (arg, arg)
    endif = "#endif"

    lines.insert(0, guard_line)
    lines.append(endif)


    return lines

def process_cpp_file(tmp_dir, arg):
    name = arg.rstrip(".cpp")
    fname = os.path.join(tmp_dir, "%s.cpp" % name)
    ofname = os.path.join(tmp_dir, "%s.o" % name)
    hfname = os.path.join(tmp_dir, "%s.h" % name)

    with open(arg) as f:
        lines = f.readlines()

    basedir = os.path.dirname(fname)

    try:
        os.makedirs(basedir)
    except:
        pass

    with open(arg) as f:
        lines = f.readlines()

    with open(fname, "w") as f:
        f.write("\n".join(lines))

    return ofname


def process_cpy_file(args, tmp_dir, arg, use_headers=False):
    name = arg.rstrip(".cpy")
    fname = os.path.join(tmp_dir, "%s.cpp" % name)
    ofname = os.path.join(tmp_dir, "%s.o" % name)
    hfname = os.path.join(tmp_dir, "%s.h" % name)

    lines = process_file(arg)
    header = analysis.extract_header(lines)
    header = add_guards(arg, header)

    basedir = os.path.dirname(fname)

    try:
        os.makedirs(basedir)
    except:
        pass

    if use_headers:
        lines = analysis.remove_structs_and_classes(lines)
        lines.insert(0, '#include "%s"' % os.path.basename(hfname))
        with open(hfname, "w") as f:
            f.write("\n".join(header))

    if (args.print_):
        print("// %s" % arg)
        print_lines(lines)
        return

    with open(fname, "w") as f:
        f.write("\n".join(lines))



def process_files(tmp_dir, args):
    files = args.files
    # if we have multiple files, we have to generate their headers
    use_headers = len(files) > 1

    for arg in files:
        if arg == '-':
            lines = sys.stdin.readlines()
            lines = pipeline.pipeline(lines)
            print_lines(lines)
        else:
            util.verbose("processing", arg)
            if arg.endswith(".cpy"):
                process_cpy_file(args, tmp_dir, arg, use_headers)
            if arg.endswith(".cpp"):
                process_cpp_file(tmp_dir, arg)
            if arg.endswith(".h"):
                process_h_file(tmp_dir, arg)

def compile_files(tmp_dir, args):
    cur_dir = os.getcwd()
    outname = args.exename or "./a.out"
    if outname[0] != '/':
        outname = os.path.join(cur_dir, outname)

    files = args.files
    more_than_stdin = False
    for arg in files:
        if arg != '-':
            more_than_stdin = True

    ofiles = []
    for arg in files:
        if arg != '-':
            if arg.endswith(".cpy"):
                ofiles.append(compile_cpy_file(tmp_dir, arg))
            if arg.endswith(".cpp"):
                ofiles.append(compile_cpp_file(tmp_dir, arg))


    if not args.print_ and more_than_stdin and not args.noexe:
        os.chdir(tmp_dir)
        util.verbose("generating", outname)
        cmd_args = ofiles + [ "-o", outname ] + COMPILE_FLAGS
        run_cmd("g++", cmd_args)

    if config.RUN_EXE:
        if config.RUN_WITH_INPUT:
            output = run_cmd(outname, stdin=sys.stdin.read())
        else:
            output = run_cmd(outname)
        util.debug('OUTPUT:\n')
        util.debug(output.decode("utf-8"))


# we need a two pass compilation so we correctly build
# all necessary header files before compiling
COMPILE_FLAGS=""
def compile_project(args):
    global COMPILE_FLAGS
    flags = []
    files = []
    for file in args.files:
        if file[0] == '-' and file != '-':
            flags.append(file)
        else:
            files.append(file)

    COMPILE_FLAGS = flags

    if COMPILE_FLAGS:
        util.debug("compile flags:", " ".join(COMPILE_FLAGS))

    if args.dir:
        tmp_dir = os.path.abspath(args.dir)
        os.makedirs(tmp_dir)
    else:
        tmp_dir = tempfile.mkdtemp()
    util.verbose("working tmp dir is", tmp_dir)


    try:
        process_files(tmp_dir, args)

        if not (args.print_):
            ofiles = compile_files(tmp_dir, args)
    finally:
        if not config.KEEP_DIR and not args.dir:
            util.verbose("removing", tmp_dir)
            shutil.rmtree(tmp_dir)
        else:
            util.debug("compiled into", tmp_dir)



from ..util import *
import os

def replace_raw(lines, base_dir):
    new_lines = []
    for line in lines:
        cline = line.strip()
        if cline.startswith("#raw "):
            args = smart_split(cline[len("#raw "):], '"')

            for arg in args:
                fname = os.path.join(base_dir, arg)
                with open(fname) as f:
                    new_lines.extend(f.readlines())
        else:
            new_lines.append(line)


    return new_lines

def replace_blocks(lines):
    new_lines = []
    for line in lines:
        if ignore_line(line, new_lines):
            continue

        indent = get_indent(line)
        cline = line.strip()
        if cline.startswith('block:'):
            line = "%s/* %s */" % (' ' * indent, cline)

        new_lines.append(line)
    return new_lines

def replace_knowns(lines):
    new_lines = []
    for line in lines:
        if ignore_line(line, new_lines):
            continue

        indent = get_indent(line)
        cline = line.strip()
        if cline.startswith('known '):
            line = "%s// %s" % (' ' * indent, cline)

        new_lines.append(line)
    return new_lines

def replace_tabs(lines):
    return [ line.replace('\t', '    ') for line in lines ]

def replace_pass(lines):
    new_lines = []
    for line in lines:
        if ignore_line(line, new_lines):
            continue

        if line.strip() == "pass":
            line = line.replace("pass", "(void)0")

        new_lines.append(line)

    return new_lines

def replace_self(lines):
    new_lines = []
    for line in lines:
        if line.find("self.") != -1:
            line = line.replace("self.", "this->")

        new_lines.append(line)

    return new_lines

def replace_loop(line, keyword='for', op='<', inc='++'):

    sline = line.strip()
    kw = "%s " % (keyword.strip())
    if sline.startswith(kw) and sline.find(";") == -1:
        # check if we are in a range loop
        range_loop = False
        if sline.find(':') != -1 and sline.find(':') != len(sline) - 1:
            range_loop = True

        if not range_loop:
            rem = sline[len(kw):].rstrip(':')
            args = smart_split(rem, ' ')

            ind = ' ' * get_indent(line)

            if len(args) == 2:
                line = "%sfor auto %s = 0; %s %s %s; %s%s" % (ind, args[0], args[0],
                    op, args[1], args[0], inc)
            if len(args) == 3:
                line = "%sfor auto %s = %s; %s %s %s; %s%s" % (ind, args[0], args[1], args[0],
                    op, args[2], args[0], inc)
            if len(args) == 4:
                line = "%sfor auto %s = %s; %s %s %s; %s += %s" % (ind, args[0], args[1], args[0],
                    op, args[2], args[0], args[3])


    return line

def replace_for_shorthand(lines):
    new_lines = []
    for line in lines:
        if ignore_line(line, new_lines):
            continue

        if config.ENABLE_FOR:
           line = replace_loop(line, keyword='for')

        if config.ENABLE_ROF:
            line = replace_loop(line, keyword='rof', op='>=', inc='--')



        new_lines.append(line)

    return new_lines

# finds and replaces "def" in front of functions
def replace_defs(lines):
    new_lines = []
    for line in lines:
        if ignore_line(line, new_lines):
            continue

        cline = line.strip()
        if cline.startswith("def "):
            tokens = cline.split()
            next_word = tokens[1]
            if next_word.find("(") == -1 or cline.find("main(") != -1 or cline.find("__new__") != -1 or cline.find("__del__") != -1:
                line = line.replace("def ", "")
            else:
                # is a function
                line = line.replace("def ", "auto ")

        new_lines.append(line)

    return new_lines




from .util import *

from . import id_recognizer
def read_scopings(lines):
    indent_levels = [0]
    nb = 0
    scopings = {}
    scope = {}
    scope_stack = { 0: scope }


    for i, line in enumerate(lines):
        line = line.rstrip('\n')
        indent = get_indent(line)

        if indent in scope_stack:
            scope = scope_stack[indent]
        else:
            scope = {}
            scope_stack[indent] = scope

        scopings[i] = [c for c in scope]

        if not line:
            continue

        if visibility_line(line):
            continue

        # last non blank line is now this one
        nb = i

        if indent_levels[-1] > indent:
            while indent_levels[-1] > indent:
                indent_levels.pop()
                scope = scope_stack[indent_levels[-1]]


        if indent_levels[-1] < indent:
            if not visibility_line(lines[nb]):
                indent_levels.append(indent)

                scope = dict([(c, c) for c in scope])
                scope_stack[indent] = scope

        if i < len(lines) - 1:
            x = 1
            # skip empty-ish lines when figuring out next indent
            while i+x < len(lines) -1 and not lines[i+x].strip() :
                x += 1

            next_line = lines[i+x]
            next_indent = get_indent(next_line)

            if indent < next_indent:
                scope = dict([(c, c) for c in scope])
                scope_stack[next_indent] = scope

        new = id_recognizer.add_identifiers(line, scope)

        # compare libclang vs. our own version
        # import clanger
        # new2 = clanger.add_identifiers(line, bscope)
        # debug("TOKENS", line, new, new2)


    return scopings

# after line is already decorated, we can check if its a function
# so we can make a .h file for this .cpp file
CONTROL_WORDS = [
    'else',
    'if',
    'class',
    'struct',
    'typedef',
    'for',
    'while']

def line_is_function(line):
    args = smart_split(line, ' ')
    if not args:
        return

    if args[0] in CONTROL_WORDS:
        return False

    if args[-1].endswith('{') and line.find('(') != -1:
        return True

    return False


def extract_header(lines):
    extracted = []
    i = 0
    while i < len(lines):
        line = lines[i]

        if line_is_include(line) or line_is_using(line):
            extracted.append(line)

        if line_is_function(line):
            func_decl = line.rstrip('{')
            func_decl += ';'
            extracted.append(func_decl)


        if is_struct(line) or is_class(line):
            until = extract_until_close(lines, i)
            ex = lines[i:until]
            extracted.extend(ex)
            i = until
        else:
            i += 1

    return extracted

def remove_structs_and_classes(lines):
    new_lines = []
    i = 0
    while i < len(lines):
        line = lines[i]
        if is_struct(line) or is_class(line):
            until = extract_until_close(lines, i)
            i = until
        else:
            i += 1
            new_lines.append(line)

    return new_lines

def guess_required_files(lines):
    requires = set()

    REQUIRE_KEYWORDS = {
        "<iostream>" : [ "cout", "cin", "endl" ],
        "<vector>" : [ "vector" ],
        "<tuple>" : [ "tuple", "make_tuple", "tie", "std::tie"],
        "<queue>" : [ "queue" ],
        "<deque>" : [ "deque" ],
        "<map>" : ["map"],
        "<unordered_map>" : ["unordered_map"],
        "<cstdio>" : ["printf", "scanf"]
    }


    def add_requires(line_tokens, require_tokens, require):
        if not require in requires:
            for tok in require_tokens:
                if tok in tokens:
                    requires.add(require)


    using_namespace_std = False
    for line in lines:
        if not using_namespace_std and line.find("using namespace std;") != -1:
            using_namespace_std = True

        for require in REQUIRE_KEYWORDS:

            tokens = smart_split(line, " :<>()")

            add_requires(tokens, REQUIRE_KEYWORDS[require], require)


    prepend = []
    for r in requires:
        prepend.append("#include %s" % r)

    included = set()
    for line in lines:
        for p in prepend:
            if line.find(p) != -1:
                included.add(p);

    prepend = list(filter(lambda w: w not in included, [ p for p in prepend]))
    if not using_namespace_std and requires:
        prepend.append("using namespace std;");

    prepend.sort()

    return prepend




from __future__ import print_function

import sys
import os

from . import config

def get_parser():
    import argparse

    parser = argparse.ArgumentParser(description='Process .cpy files into C++')
    parser.add_argument('-ni', '--disable-implication', help='disables variable implication', action='store_true')
    parser.add_argument('files', nargs='*', help="list of files to process and compile")
    parser.add_argument('-r', '--run', dest="runexe", action="store_true",
        help="invoke executable after compiling it")
    parser.add_argument('-ri', '--run-with-stdin', dest="runinput", action="store_true",
        help="invoke executable after compiling it and pass stdin to it")
    parser.add_argument('-o', '--output', dest='exename', default="a.out",
        help="set the name of the executable file")
    parser.add_argument('-d', '--dir', dest='dir', default=None,
        help="output source files to this directory. implies -k")
    parser.add_argument('-v', '--verbose', dest='verbose', action="store_true")
    parser.add_argument('-p', '--print', dest='print_', action="store_true",
        help="print translated C++ source code")
    parser.add_argument('-k', '--keep-dir', dest='keep_dir', action="store_true",
        help="keep compilation directory around")
    parser.add_argument('-c', '-ne', '--no-exe', dest='noexe', action="store_true",
        help='compile .o files only (no main)')
    parser.add_argument('-for', '--enable-for', dest='enable_for', action="store_true",
        help="enable for loop shorthand")
    parser.add_argument('-rof', '--enable-rof', dest='enable_rof', action="store_true",
        help="enable rof loop shorthand")

    return parser

def main():
    parser = get_parser()
    args = parser.parse_args()
    if not args.files:
        parser.print_help()
        return

    if args.disable_implication:
        transforms.variables.DECLARE_VARIABLES = False

    config.VERBOSE = args.verbose

    config.ENABLE_ROF = args.enable_rof
    config.ENABLE_FOR = args.enable_for

    config.KEEP_DIR = args.keep_dir
    config.RUN_EXE = args.runexe
    config.RUN_WITH_INPUT = args.runinput
    if args.runinput:
        config.RUN_EXE = True

    from .project import compile_project
    compile_project(args)

if __name__ == "__main__":
    main()
